/*
*  C Implementation: channelData.c
*
* Description: A class to support update of vdr epg events
*
*
* Author: RealQuick <realquick@doublequick.de>, (C) 2009
*
* Copyright: See COPYING file that comes with this distribution
*
*/

#include "channelData.h"
#include <vdr/channels.h>       // for Channels
#include <vdr/epg.h>
#include <iostream>

#include "pngutils.h"
#include "common.h"

#include "iepg.h"

#include "svdrpc.h"
#include "setup.h"


#define BUFFSIZE 32000


using namespace std;


cChannelData::cChannelData ( const char *dir, const char *epgFileName, int ProviderType )
{
    isLoaded = false;
    EpgDataFileName = "";
    EpgDir = "";

    EpgDataFileName =  epgFileName;
    EpgDir = dir;

    m_ProviderType = ProviderType;

    isLoaded = true;

    tipdesc[0] = NULL;
    tipdesc[1] = tr("Tip of the day");
    tipdesc[2] = tr("Tip");
    tipdesc[3] = tr("Top tipp");

    return;
}                               //end cChannelData(char*)

cChannelData::~cChannelData ()
{
    ProgrammeList.clear();
}

int cCleanChannelData::ClearEvents(cSchedule *Schedule, tChannelID cid)
{
    std::vector < cEvent *> EventsToDelete;

    cChannel *chn = Channels.GetByChannelID(cid);
    if (chn)
        dsyslog("iepg: About to clean Channel %s", chn->Name());


    cEvent *p = NULL;
    cEvent *EventToMove = NULL;
    const cEvent *last_Event = NULL;
    const cEvent *next_Event = NULL;
    int nDeleted = 0;

    bool bMoveEvent;

    const cSchedules *tmp_schedules = NULL;
    cSchedule *tmp_schedule = NULL;

    cSchedulesLock *upd_schedulesLock = NULL;

    if (!Schedule)
    {
        cSchedulesLock* ReadLock = new cSchedulesLock ( false, 100 );
        tmp_schedules = cSchedules::Schedules ( *ReadLock );
        if (!tmp_schedules)
        {
            dsyslog("iepg: No Channel data available!");
            return 0;
        }
        else
            tmp_schedules = NULL;

        delete ReadLock;

        upd_schedulesLock = new cSchedulesLock ( true, 180000 );	// Wait max 3 Min for freeing of Lock
        if (!upd_schedulesLock)
        {
            esyslog("iepg: new cSchedulesLock FAILED");
            return 0;
        }

        tmp_schedules = cSchedules::Schedules ( *upd_schedulesLock );
        if (!tmp_schedules)
        {
            delete upd_schedulesLock;
            esyslog("iepg: cSchedules::Schedules ( upd_schedulesLock ) FAILED");
            return 0;
        }

        tmp_schedule = ( cSchedule * ) tmp_schedules->GetSchedule ( cid ); // channelID id
    }
    else
        tmp_schedule = Schedule;

    // No Channel
    if ( !tmp_schedule )
    {
        if (!upd_schedulesLock)
        {
            delete upd_schedulesLock;
            upd_schedulesLock = NULL;
        }
        dsyslog("iepg: NO channel data!");
        return(0);
    }

    // No events in channel
    if (!tmp_schedule->Events())
    {
        if (!upd_schedulesLock)
            delete upd_schedulesLock;
        dsyslog("iepg: No events in channel!");
        return(0);
    }

    const cList <cEvent> *events = tmp_schedule->Events();
    if (!events)
    {
        if (!upd_schedulesLock)
            delete upd_schedulesLock;
        return(0);
    }

    for (p = events->First(); p /* && nDeleted < 10  TEST */; p = events->Next(p))
    {
        bMoveEvent = false;
        if (!p->HasTimer() && p->TableID())	// Tableid <> 0 -> from DVB-Stream -> Delete event
        {
            next_Event = events->Next(p);
            if (next_Event)
                if (next_Event->StartTime() - p->StartTime() < MAX_EPG_DELTA_SECONDS && next_Event->TableID() == 0)
                {
                    EventToMove = (cEvent *)next_Event;
                    bMoveEvent = true;
                }
            if (last_Event && !bMoveEvent)
                if (p->StartTime() - last_Event->StartTime() < MAX_EPG_DELTA_SECONDS && last_Event->TableID() == 0)
                {
                    EventToMove = (cEvent *)last_Event;
                    bMoveEvent = true;
                }

            if (bMoveEvent)
            {
                struct tm *ts;
                time_t t = p->StartTime();
                ts = localtime(&t);

                if (p->StartTime())
                    esyslog ("iepg: title %s(%s) starttime %s Duration: %d Min (id %ld) Event moved by %d minutes", p->Title(), EventToMove->Title(),  asctime(ts), p->Duration() / 60, (long)p->EventID(), (int)((EventToMove->StartTime() - p->StartTime()) / 60));

		int nDiffDuration =  EventToMove->StartTime() - p->StartTime();
                EventToMove->SetStartTime(p->StartTime());
                EventToMove->SetDuration(p->Duration() - nDiffDuration);

                // Mark actual event to remove
                p->SetStartTime(0);
                p->SetDuration(60);


                tmp_schedule->SetRunningStatus ( p, SI::RunningStatusNotRunning, Channels.GetByChannelID ( cid ) );
                tmp_schedule->SetRunningStatus ( EventToMove, SI::RunningStatusNotRunning, Channels.GetByChannelID ( cid ) );

           	nDeleted++;

            }
	    
	    EventsToDelete.push_back(p);

        }
        last_Event = p;
    }

    // Loop through all events marked for deletion
    for (unsigned int i=0; i<EventsToDelete.size (); i++)
    {
         tmp_schedule->UnhashEvent(EventsToDelete[i]);
         tmp_schedule->DelEvent(EventsToDelete[i]);
    }

    tmp_schedule->Sort ();
    tmp_schedule->SetModified ();
    tmp_schedules->SetModified (tmp_schedule);

    // release Lock
    if ( upd_schedulesLock != NULL )
    {
        delete upd_schedulesLock;
        upd_schedulesLock = NULL;
    }

    return nDeleted;
};

void cCleanChannelData::Action()
{
    CleanNow(true);
}

int cCleanChannelData::CleanNow(bool bThread)
{
    int nDeletedTotal = 0;

    cSchedulesLock *upd_schedulesLock;
    const cSchedules *tmp_schedules;
    cSchedule *tmp_schedule = NULL;

    if (bThread)
    {
        Lock();
    }

    // Channel id provided -> process this channel and return
    if (!(m_cid == tChannelID()))
    {
        nDeletedTotal = ClearEvents(m_Schedule, m_cid);
        if (bThread)
            Unlock();
        return nDeletedTotal;
    }

    upd_schedulesLock = new cSchedulesLock ( true, bThread ? 0 : 120000 );	// Wait max 2 Min for freeing of Lock
    if (!upd_schedulesLock)
    {
        esyslog("iepg: new cSchedulesLock FAILED");
        if (bThread)
            Unlock();
        return -1;
    }

    tmp_schedules = cSchedules::Schedules ( *upd_schedulesLock );
    if (!tmp_schedules)
    {
        delete upd_schedulesLock;
        esyslog("iepg: cSchedules::Schedules ( upd_schedulesLock ) FAILED");
        if (bThread)
            Unlock();
        return -1;
    }

    else
    {
        // Loop through all vdr channels
        // If channel is mapped to iepg -> Clean channel from dvb-Events
        for (const cChannel *channel = Channels.First(); channel && channel->Number() < MAX_VDR_CHANNELS; channel = Channels.Next(channel))
        {
            if (m_Reverse_ChannelList[channel->Number()])
            {
//                cChannel *chn = Channels.GetByNumber(i);
                if (channel)
                {
                    tmp_schedule = ( cSchedule * ) tmp_schedules->GetSchedule ( channel->GetChannelID() ); // channelID id
                    if (tmp_schedule)
                    {
                        int nDeleted = ClearEvents(tmp_schedule, channel->GetChannelID());
                        if (nDeleted)
                            dsyslog("iepg: Channel %s : %d Events from DVB removed", channel->Name(), nDeleted);

			nDeletedTotal += nDeleted;
                    }
                    else
                        dsyslog("iepg: No channel data");
                }
                delete upd_schedulesLock;

                upd_schedulesLock = new cSchedulesLock ( true, bThread ? 0: 120000 );	// Wait max 2 Min for freeing of Lock
                if (!upd_schedulesLock)
                {
                    esyslog("iepg: new cSchedulesLock FAILED");
                    if (bThread)
                        Unlock();
                    return -1;
                }

                tmp_schedules = cSchedules::Schedules ( *upd_schedulesLock );
                if (!tmp_schedules)
                {
                    delete upd_schedulesLock;
                    esyslog("iepg: cSchedules::Schedules ( upd_schedulesLock ) FAILED");
                    if (bThread)
                        Unlock();
                    return -1;
                }

            }
        }
    }

    if (upd_schedulesLock)
        delete upd_schedulesLock;

    if (bThread)
        Unlock();

    return nDeletedTotal;
};

cCleanChannelData::cCleanChannelData(cSchedule *Schedule, tChannelID cid, int * Reverse_ChannelList)  :cThread ( "Cleanup duplicate EPG-Entries" )
{
    m_Schedule = Schedule;
    m_cid = cid;
    m_Reverse_ChannelList = Reverse_ChannelList;
}

// int cChannelData::loadEpgData (const char *channelName, char *channelId)
int cChannelData::loadEpgData ( cProviderChannelList *ChannelList, int nChannelID)
{
    /*
     *  loads EPG data into Channel "channelName"
     */

    if ( !parseEpgDataFile ( ChannelList ) )
        return -1;                  // parsing failed

    cSchedulesLock *tmp_schedulesLock = NULL;
//    cSchedule *tmp_schedule = NULL;

    int prgCount = 0;             // programme count
    // prgCount is returned

    tChannelID cid;
    char crc_str[256];


    for ( int nIndex=(nChannelID >= 0 ? nChannelID : 0); nIndex <=  (nChannelID >= 0 ? nChannelID : (MAX_PROVIDER_CHANNEL + PROVIDER_HASHSIZE + COLLISION_TABSIZE - 1)); nIndex++ ) // Loop through channels - if nChannelID is provided only 1 time
    {
        if ( ChannelList[nIndex].ChannelId.Source() )  // Channel data is available -> load it
        {

            /* tChannelID */
            cid = ChannelList[nIndex].ChannelId;

	    cSchedules *tmp_schedules = NULL;
	    while ( tmp_schedules == NULL )
	    {
            	if ( tmp_schedulesLock == NULL )
               	    tmp_schedulesLock = new cSchedulesLock ( true, 60000 );

            	tmp_schedules = ( cSchedules * ) cSchedules::Schedules ( *tmp_schedulesLock );
	    	if( tmp_schedules == NULL)
		{
                       esyslog("Can not lock Schedules! Will try again later." );
                       	do_log ( iepgSetup.sDataDirectory.c_str(), LOGFILE_THREAD, tr("EPG data is currently locked by another program! Will try again later.") );
			delete tmp_schedulesLock;
			tmp_schedulesLock = NULL;
			sleep(60);	// Wait again 60 seconds and try once more
		} else
		{
			if( !tmp_schedulesLock->Locked())
			{
                        	esyslog("EPG data is currently locked by another program! Will try again later." );
                        	do_log ( iepgSetup.sDataDirectory.c_str(), LOGFILE_THREAD, tr("EPG data is currently locked by another program! Will try again later.") );
				delete tmp_schedulesLock;
				tmp_schedulesLock = NULL;
				delete tmp_schedules;
				tmp_schedules = NULL;
				sleep(60);	// Wait again 60 seconds and try once more
			}
		}
	    }

	    cSchedule *tmp_schedule;
	    if(tmp_schedules)
            	tmp_schedule = ( cSchedule * ) tmp_schedules->GetSchedule ( cid ); // channelID id

            /*
             * if no schedule exists for the channel
             * create a new one
             */
            if ( !tmp_schedule )
            {
                tmp_schedule = new cSchedule ( cid );
                tmp_schedules->Add ( tmp_schedule );
                dsyslog ( "iepg: create new schedule for channel %d", cid.Source() );
            }
		
	    if(!tmp_schedules)
                esyslog ( "iepg: Could not get Schedules! Can not import epg data at this time");

            int duration, daysForward;
            time_t now, lastEndTime;
            time_t StartTime, EndTime;
            time_t VpsTime;
//            char crc_str[256];
            tEventID tmp_eventID;

            std::vector < programmeData >::iterator iter = ProgrammeList.begin ();

//         iepgStartMessage(EPG_UPDATE, tr("update channels"));

            for ( ; tmp_schedules && iter != ProgrammeList.end (); iter++ )
            {
                duration = daysForward = 0;
                now = lastEndTime = 0;
                StartTime = EndTime = 0;
                VpsTime = 0;

                // only process entries from the given channel (bleb.org has only one channel per xml)
                if ( nChannelID > 0 || iter->ChannelID == cid )
                {
                    int emptyEndTime = 0;
                    now = time ( NULL );

                    if ( strcmp ( iter->end_time.c_str (), "" ) == 0 )
                    {
                        emptyEndTime = 1;     //TODO log error
                    }

                    switch ( m_ProviderType )
                    {
                    case 0:
//FIXME
                        StartTime = str2time_t ( iter->start_time.c_str (), iter->date.c_str (), daysForward );
                        if ( !emptyEndTime )
                            EndTime = str2time_t ( iter->end_time.c_str (), iter->date.c_str (), daysForward );

                        /*
                         * Correct for "improper" Times
                         * like change of day, indicated just by time and not date
                              *
                              * eg. 2330-0015 programme Title
                             */

                        if ( !emptyEndTime )
                        {
                            if ( ( double ) lastEndTime > ( double ) EndTime )
                            {
                                // StartTime before midnight and EndTime after midnight
                                EndTime += 24 * 60 * 60;
                                daysForward++;      // assumes date.c_str() doesnot change
                            }
                        }

                        if ( ( double ) lastEndTime > ( double ) StartTime )
                        {
                            // assume the schedule points to next day's event
                            // add 24*60*60 to both EndTime and StartTime
                            EndTime += 24 * 60 * 60;
                            StartTime += 24 * 60 * 60;
                            daysForward++;
                        }

                        break;
                    case 1:
                        cChannel *chan = Channels.GetByChannelID(cid);
                        if (iter->vps != "")
                        {
                            dsyslog("iepg: VPS: Channel %s, VPS-Time: %s (%s)", chan->Name(), iter->vps.c_str (), (iter->start_time.substr(0, 11) + iter->vps + ":00").c_str());
                            VpsTime = dtstr2time_t ((iter->start_time.substr(0, 11) + iter->vps + ":00").c_str());
                        }
                        StartTime = dtstr2time_t ( iter->start_time.c_str () );
                        if ( !emptyEndTime )
                            EndTime = dtstr2time_t ( iter->end_time.c_str () );

                        if ( ( double ) lastEndTime > ( double ) StartTime )
                        {
                            dsyslog("iepg: midnight bug: Channel %s, Starttime: %s", chan->Name(), iter->start_time.c_str());
                        }

                        break;
                    }

                    if ( !emptyEndTime )
                    {
                        duration = ( int ) difftime ( EndTime, StartTime );
                        lastEndTime = EndTime; // to take care when schedule spills into the following day
                    }

                    struct tm *ts;
                    ts = localtime ( &StartTime );

                    /* CORRECT FOR TIME ZONE
                     * assume that programme times are in GMT+1
                     *
                     * use tzset() to find your TimeZone
                     * add timezone (in sec) to startTime and endTime
                     *
                     * TODO check if StartTime and EndTime are either side of DST/timezone change.
                     */

                    // adjust timezone only for provider 0 (BBC SKY)
                    //TODO check timezones for other countries, different from germany
                    if ( m_ProviderType == 0 )
                    {
                        struct tm *local;
                        int nTimeOffset = 0;

                        tzset();
                        local = localtime(&StartTime);
                        nTimeOffset = timezone;

                    	// does daylight saving takes place?
                        if (local->tm_isdst)
                            nTimeOffset += 60*60;

                        StartTime -= nTimeOffset;
                        EndTime -= nTimeOffset;
                    }

                    cEvent *Event = NULL;

                    /*
                     *  calculate a unique eventID
                     *  from title and startTime
                     */
//                    sprintf ( crc_str, "%s%32lf", iter->title.c_str (), ( double ) StartTime );
//                    tmp_eventID = ( ( ( ( 'P' << 8 ) | 'W' ) << 16 ) | ( crc16 ( 0, ( unsigned char * ) crc_str, sizeof ( crc_str ) ) ) ) & 0xFFFF;

                    // 32 bit unique number
                    if ( iter->eventid )
                        tmp_eventID = iter->eventid;
                    else
                    {
                        sprintf (crc_str, "%s%32lf", iter->title.c_str (), (double) StartTime);
                        tmp_eventID = (((('P' << 8) | 'W') << 16) | (crc16 (0, (unsigned char *) crc_str, sizeof (crc_str)))) & 0xFFFF;
		    }

                    Event = ( cEvent * ) tmp_schedule->GetEvent ( ( tEventID ) NULL, StartTime );
                    if ( Event )
                    {
                        if ( ( Event->EventID() != tmp_eventID ) )
                        {
                            if ( !Event->HasTimer() )
                            {
                                dsyslog ( "iepg: remove event from foreign source (dvb?) [%s]", Event->Title() );
                                tmp_schedule->SetPresentSeen();
                                tmp_schedule->UnhashEvent ( Event );
                                tmp_schedule->DelEvent ( Event );
                                Event = NULL;
                                dsyslog ( "iepg: event from foreign source removed" );
                            };
                        }
                    }

                    cEvent *newEvent = NULL;
                    if ( !Event )
                    {
                        Event = newEvent = new cEvent ( tmp_eventID );
                    }

                    //FIXME only new events added. events from foreign sources (dvb) remains untouched
                    if ( newEvent )
                    {
                        // if pictures available move file to /opt/iepg/images/eventid.png
                        if ( iter->png.size() > 0 )
                            if ( file_exists ( ( char * ) iter->png.c_str() ) )
                            {
                                char cmd[256];
                                sprintf ( cmd, "ln %s %s/%ld.png", iter->png.c_str(), iepgSetup.sImageDirectory.c_str(), ( long ) Event->EventID() );
                                SystemExec ( cmd );

                                string r = iter->png;
                                r.erase ( iter->png.size() - 4 );
                                r += ".raw";

                                /*
                                    if(unlink(cmd))
                                     esyslog("iepg: could not unlink %s", cmd);
                                */
                                if ( file_exists ( ( char * ) r.c_str() ) )
                                {
                                    touch_file ( ( char * ) r.c_str(), StartTime );
                                    sprintf ( cmd, "ln %s %s/%ld.raw", r.c_str(), iepgSetup.sImageDirectory.c_str(), ( unsigned long ) Event->EventID() );
                                    SystemExec ( cmd );

                                    do_log ( EpgDir.c_str(), LOGFILE_IMPORT, cmd );
                                }

                                /* Setting times of symblic link does not work
                                   // Set creation date to startdate of event
                                                       sprintf ( pic, "%s/%ld.raw", iepgSetup.sImageDirectory, ( long ) EventID );
                                   touch_file(pic, StartTime);
                                */
                            }


                        /*
                         * set Title, Short Text, Description, start and end time etc...
                         */


                        // FIXME workaround for epgdata.com special characters
                        // Title: Replace em dash  (0xc297) into "-"
                        for (unsigned int x=0;x<(iter->title.size() - 1);x++) if (iter->title.at(x) == (char)0xc2  && iter->title.at(x+1) == (char)0x97) iter->title.replace(x, 2, "-");

                        // Replace linefeeds into spaces (problems during read / write to file epg.data)
                        for (unsigned int x=0;x<iter->title.size();x++) if (iter->title.at(x) == (char)0x10) iter->title.replace(x, 1, " ");
                        for (unsigned int x=0;x<iter->short_text.size();x++) if (iter->short_text.at(x) == (char)0x10) iter->short_text.replace(x, 1, " ");
                        for (unsigned int x=0;x<iter->description.size();x++) if (iter->description.at(x) == (char)0x10) iter->description.replace(x, 1, " ");

                        Event->SetTitle ( iter->title.c_str () );
                        Event->SetShortText ( iter->short_text.c_str () );

                        Event->SetDescription ( iter->description.c_str () );

                        Event->SetStartTime ( StartTime );
                        if ( !emptyEndTime )
                            Event->SetDuration ( duration );

                        if ( VpsTime )
                            Event->SetVps ( VpsTime );

                        int runningStatus = 0;
                        if ( !emptyEndTime )
                            runningStatus = ( StartTime < now &&
                                              now < EndTime ) ? SI::RunningStatusRunning :
                                            ( ( StartTime - 30 < now && now < StartTime ) ? SI::RunningStatusStartsInAFewSeconds : SI::RunningStatusNotRunning );

                        Event->SetRunningStatus ( runningStatus, Channels.GetByChannelID ( cid ) );
                        Event->SetVersion ( 0xFF );
                        Event->SetTableID ( 0x0 ); //0x2E   // when 0x0 it is not modified by the EPG got from DVB broadcasted

                        Event->SetComponents ( NULL );
                        Event->FixEpgBugs ();

                        const cEvent *tmp_delEvent = tmp_schedule->GetEventAround ( StartTime );
                        if ( tmp_delEvent )     // remove this event
                        {
                            //time_t eT = tmp_delEvent->EndTime (), sT = tmp_delEvent->StartTime ();
                            tmp_schedule->DropOutdated ( StartTime, EndTime + 60, Event->TableID (), !tmp_delEvent->Version () );
                        }

                        tmp_schedule->AddEvent ( Event );
                        tmp_schedule->SetRunningStatus ( Event, runningStatus, Channels.GetByChannelID ( cid ) );
                        tmp_schedule->Sort ();

                        tmp_schedule->SetModified ();
                        tmp_schedules->SetModified ( tmp_schedule );

                        prgCount++;
                    }
                }                         // end  programmeList loop

            }                           // endif comparing c->Name()

            // If noEPG-Patch is not active -> drop/move/replace duplicate entries
            if (iepgSetup.NoEPGWorkaround != NOEPG_PATCHED_VDR)
            {
                cCleanChannelData clean(tmp_schedule);
                //       int nClear = ClearEvents(tmp_schedule);
                int nClear = clean.ClearEvents(tmp_schedule);
                if (nClear)
                    dsyslog("iepg: %d foreign (DVB) Events where removed from schedule", nClear);
            }
            // release Lock
            if ( tmp_schedulesLock != NULL )
            {
                delete tmp_schedulesLock;
                tmp_schedulesLock = NULL;
            }
        }
    }                             // end channel loop


    ProgrammeList.clear();

    return prgCount;
}                               // end LoadEpgData(char*)

bool
cChannelData::parseEpgDataFile ( cProviderChannelList *ChannelList )
{
    string sPath;

    if ( !isLoaded )
        return false;

    ProgrammeList.clear(); // Removes all entries from previous channels

    switch ( m_ProviderType )
    {
    case 0: /* Handle xml document for Provider 0 */
#ifdef USE_TINY_XML 
      /*
        	    sPath = EpgDir;
        	    sPath += "/";
        	    sPath += EpgDataFileName;
        */
        doc.LoadFile(EpgDataFileName.c_str());

        if ( getChannelData_Prov0 ( &doc /* , this */ ) == -1 )
        {
            return false;
        }
        break;
#else
	return false;
#endif
    case 1: /* Handle xml document for Provider 1 */

//            if ( getChannelData_Prov1 ( doc, /* this, */ ChannelList ) == -1 )
        if ( getChannelData_Prov1 ( ChannelList ) == -1 )
        {
            return false;
        }
        break;
    }

    return true;

}                               // end parseEpgDataFile()

#ifdef USE_TINY_XML
/* Channeldata from Provider bleb.org */
int cChannelData::getChannelData_Prov0 ( TiXmlNode * pParent /* , cChannelData * ch_data */ )
{

    if ( !pParent )
        return -1;                  // return an error flag

    TiXmlNode *pChild;
    TiXmlText *pText;
    int t = pParent->Type ();

    programmeData prog_data;


    switch ( t )
    {
    case TiXmlNode::DOCUMENT:
        break;

    case TiXmlNode::ELEMENT:
        if ( strcmp ( pParent->Value (), "channel" ) == 0 )
        {
            if ( pParent->ToElement () == NULL )
            {
                esyslog ( "iepg: getChannelData() Error: No channel data" );
            }
            else
            {
                TiXmlAttribute *pAttrib = pParent->ToElement ()->FirstAttribute ();
                while ( pAttrib )
                {
                    if ( strcmp ( pAttrib->Name (), "id" ) == 0 )
                    {
                        id = pAttrib->Value ();
                    }
                    if ( strcmp ( pAttrib->Name (), "source" ) == 0 )
                    {
                        source = pAttrib->Value ();
                    }
                    if ( strcmp ( pAttrib->Name (), "date" ) == 0 )
                    {
                        date = pAttrib->Value ();
                    }
                    pAttrib = pAttrib->Next ();
                }
            }
        }
        else if ( strcmp ( pParent->Value (), "programme" ) == 0 )
        {
            // get programme data
            for ( pChild = pParent->FirstChild (); pChild != 0; pChild = pChild->NextSibling () )
            {
                // get programme data
                if ( pChild->Type () == TiXmlNode::ELEMENT )
                {
                    /* TODO
                     * for (iterate over a string list)
                     *   if (strcmp(pChild->Value(), iter[i].cstring)==0)
                     *      if ( pChild->FirstChild()!=NULL && (pText = pChild->FirstChild()->ToText())!=NULL)
                     *                tmp_str =pChild->FirstChild ()->ToText () -> Value(); else tmp_str=""
                     *                switch(i) { case 1: prog_date.title = tmp_str; break; case 2:... }
                     */
                    if ( strcmp ( pChild->Value (), "title" ) == 0 )
                    {
                        if ( pChild->FirstChild () != NULL && ( pText = pChild->FirstChild ()->ToText () ) != NULL )
                        {
                            prog_data.title = pChild->FirstChild ()->ToText () -> Value ();
                        }
                        else
                        {
                            prog_data.title = "";
                        }
                    }
                    if ( strcmp ( pChild->Value (), "desc" ) == 0 )
                    {
                        if ( pChild->FirstChild () != NULL && ( pText = pChild->FirstChild ()->ToText () ) != NULL )
                        {
                            prog_data.description = pChild->FirstChild ()->ToText () -> Value ();
                        }
                        else
                        {
                            prog_data.description = "";
                        }
                    }
                    if ( strcmp ( pChild->Value (), "end" ) == 0 )
                    {
                        if ( pChild->FirstChild () != NULL && ( pText = pChild->FirstChild ()->ToText () ) != NULL )
                        {
                            prog_data.end_time = pChild->FirstChild ()->ToText () -> Value ();
                        }
                        else
                        {
                            prog_data.end_time = "";
                        }
                    }
                    if ( strcmp ( pChild->Value (), "start" ) == 0 )
                    {
                        if ( pChild->FirstChild () != NULL && ( pText = pChild->FirstChild ()->ToText () ) != NULL )
                        {
                            prog_data.start_time = pChild->FirstChild ()->ToText () -> Value ();
                        }
                        else
                        {
                            prog_data.start_time = "";
                        }
                    }
                    if ( strcmp ( pChild->Value (), "date" ) == 0 )
                    {

                        if ( pChild->FirstChild () != NULL && ( pText = pChild->FirstChild ()->ToText () ) != NULL )
                        {
                            prog_data.date = pChild->FirstChild ()->ToText () -> Value ();
                        }
                        else
                        {
                            prog_data.date = "";
                        }
                    }
                    if ( strcmp ( pChild->Value (), "infourl" ) == 0 )
                    {

                        if ( pChild->FirstChild () != NULL && ( pText = pChild->FirstChild ()->ToText () ) != NULL )
                        {
                            prog_data.infourl = pChild->FirstChild ()->ToText () -> Value ();
                        }
                        else
                        {
                            prog_data.infourl = "";
                        }
                    }
                }
            }
            /* ch_data-> */

            prog_data.date = date;
            prog_data.id = id;
            prog_data.source = source;

            ProgrammeList.push_back ( prog_data );
            pParent = ( TiXmlNode * ) 0; // avoid this subtree
        }
        break;

    case TiXmlNode::COMMENT:

        break;

    case TiXmlNode::UNKNOWN:

        break;

    case TiXmlNode::TEXT:


        break;

    case TiXmlNode::DECLARATION:

        break;
    default:
        break;
    }
    if ( pParent != 0 )
        for ( pChild = pParent->FirstChild (); pChild != 0; pChild = pChild->NextSibling () )
        {
            if ( getChannelData_Prov0 ( pChild /* , ch_data */ ) == -1 )
                return -1;
        }

    return 1;
}
#endif

/* Channeldata from Provider epgdata.com */
int cChannelData::getChannelData_Prov1 ( cProviderChannelList *ChannelList )
{
    cPngUtils img = cPngUtils ( 32000, 32000 );  // Images max. size 32000 x 32000 pixel
    FILE *xmlFile;

    char *line;
    string buffer;

    int t;

    string sElement;
    string sValue;

    string sFile =  EpgDir;
    sFile += "/";
    sFile += EpgDataFileName;

    xmlFile = fopen ( sFile.c_str(), "r" );
    if ( xmlFile == NULL )
    {
        //TODO error msg
        esyslog ( "iepg: error opening %s. Aborting", sFile.c_str() );
        return -1;
    }

    programmeData prog_data;
    prog_data.update = false;

    bool data_complete = false;

    line = new char[BUFFSIZE];

    while ( !feof ( xmlFile ) )
    {
        if ( fgets ( line , BUFFSIZE, xmlFile ) != NULL )
        {
            buffer = line;
            memset ( line, '\0', BUFFSIZE );

            sElement = sValue = "";

            unsigned int pos1 = buffer.find ( "<d" );
            if ( pos1 !=string::npos )
            {
                unsigned int pos2 = buffer.find ( ">", pos1 );
                if ( pos2 !=string::npos )
                {
                    sElement = buffer.substr ( pos1+1, pos2 - pos1 - 1 );

                    unsigned int pos3 = buffer.find ( "</d", pos2 );

                    if ( pos3 !=string::npos )
                        sValue = buffer.substr ( pos2+1, pos3 - pos2 - 1 );
                }
            }

            if ( sElement.size() )
            {
                int nHash;

                switch ( atoi ( sElement.substr ( 1 ).c_str() ) )
                {
                case 0:  // Tag d0 -> broadcast id
                    prog_data.eventid = atol ( sValue.c_str() );
                    break;
                case 2:  // d2 -> tv channel id
                    nHash = GetHashIndex(ChannelList, sValue.c_str());
                    if (nHash)
                    {
                        prog_data.update = true;
                        prog_data.ChannelID = ChannelList[nHash].ChannelId;
                    }
                    else
                        prog_data.update = false;
                    break;
                case 3:  // Tag d3 -> region id // causes problem with SWR due multiple events (with different region ids)
                    prog_data.region_id = atoi ( sValue.c_str() );
                    break;
                case 19: // Tag d19 -> Title
                    prog_data.title = "";
                    prog_data.title += sValue;
                    setup_utf8 ( &prog_data.title );
                    break;
                case 18:  // Tag d18 -> tip id
                    t=atoi ( sValue.c_str() );
                    if (t > 0 && t < 4)
                        prog_data.toptip = tipdesc[t];
                    else
                        prog_data.toptip = "";
                    break;
                case 8:  // Tag d8 -> VPS
                    prog_data.vps = sValue;
                    break;
                case 17:  // Tag d17 -> live id
                    prog_data.live = atoi ( sValue.c_str() );
                    break;
                case 14:  // Tag d14 -> coded
                    prog_data.coded = atoi ( sValue.c_str() );
                    break;
                case 16:  // Tag d16 -> age marker
                    prog_data.age_marker = atoi ( sValue.c_str() );
                    break;
                case 28:  // Tag d28 -> Dolby
                    prog_data.dolby = atoi ( sValue.c_str() );
                    break;
                case 27:  // Tag d27 -> Stereo
                    prog_data.stereo = atoi ( sValue.c_str() );
                    break;
                case 29:  // Tag d29 -> wide
                    prog_data.wide = atoi ( sValue.c_str() );
                    break;
                case 26:  // Tag d26 -> sequence
                    prog_data.sequence = atoi ( sValue.c_str() );
                    break;
                case 9:  // Tag d9 -> primetime
                    prog_data.primetime = atoi ( sValue.c_str() );
                    break;
                case 21: // Tag d21 -> long description
                    prog_data.description = sValue;
                    setup_utf8 ( &prog_data.description );
                    break;
                case 4:  // Tag d4 -> Start time
                    prog_data.start_time = sValue;
                    break;
                case 5:  // Tag d5 -> End time
                    prog_data.end_time = sValue;
                    break;
                case 25:  // tag d25 Genre
                    prog_data.genre = sValue;
                    setup_utf8 ( &prog_data.genre );
                    break;
                case 32:  // tag d32 country
                    prog_data.country = sValue;
                    break;
                case 37:  // tag 37 actor
                    prog_data.actor = sValue;
                    setup_utf8 ( &prog_data.actor );
                    break;
                case 33:  // tag d33 year
                    prog_data.year = sValue;
                    break;
                case 34:  // tag d34 moderator
                    prog_data.moderator = sValue;
                    setup_utf8 ( &prog_data.moderator );
                    break;
                case 35:  // tag d35 studio guest
                    prog_data.studio_guest = sValue;
                    setup_utf8 ( &prog_data.studio_guest );
                    break;
                case 36:  // tag d35 regisseur
                    prog_data.regisseur = sValue;
                    setup_utf8 ( &prog_data.regisseur );
                    break;
                case 40:  // tag d40 picture
                    prog_data.jpeg = sValue;
                    break;
                default:
                    break;
                }
                prog_data.infourl = "";
            }

            if ( buffer.find ( "</data>" ) !=string::npos )
            {
                data_complete = true;
            }
            else
                if ( buffer.find ( "<data>" ) !=string::npos )
                {
                    data_complete = false;
                }


            if ( data_complete )
            {
                if ( prog_data.update && ( prog_data.region_id == 0 ) )  // Do not import events for differerent regions e. g. SWR
                {
                    // reset flags
                    data_complete = false;
                    prog_data.update = false;

                    string stext;
                    stext += LINEFEED;
                    stext = tr ( "Genre: " );
                    stext += get_genre ( prog_data.genre );
                    stext += LINEFEED;
                    stext += LINEFEED;
                    stext += prog_data.description;
                    prog_data.description = stext;
                    stext += LINEFEED;

                    if ( prog_data.country.size() )
                    {
                        stext = LINEFEED;
                        stext += tr ( "Country: " );
                        stext += prog_data.country;
                        prog_data.description += stext;
                    }
                    if ( prog_data.actor.size() )
                    {
                        stext = LINEFEED;
                        stext += tr ( "Actor: " );
                        stext += prog_data.actor;
                        prog_data.description += stext;
                    }
                    if ( prog_data.year.size() )
                    {
                        stext = LINEFEED;
                        stext += tr ( "Year: " );
                        stext += prog_data.year;
                        prog_data.description += stext;
                    }
                    if ( prog_data.regisseur.size() )
                    {
                        stext = LINEFEED;
                        stext += tr ( "Regisseur: " );
                        stext += prog_data.regisseur;
                        prog_data.description += stext;
                    }
                    if ( prog_data.moderator.size() )
                    {
                        stext = LINEFEED;
                        stext += tr ( "Moderator: " );
                        stext += prog_data.moderator;
                        prog_data.description += stext;
                    }
                    if ( prog_data.studio_guest.size() )
                    {
                        stext = LINEFEED;
                        stext += tr ( "Studio guest: " );
                        stext += prog_data.studio_guest;
                        prog_data.description += stext;
                    }

                    if ( prog_data.toptip.size() )
                    {
                        stext = LINEFEED;
                        stext += tr ( "Toptip: " );
                        stext += prog_data.toptip;
                        prog_data.description += stext;
                    }
                    if ( prog_data.live )
                    {
                        stext = LINEFEED;
                        stext += tr ( "Live event: " );
                        stext += tr ( "yes" );
                        prog_data.description += stext;
                    }
                    if ( prog_data.coded )
                    {
                        stext = LINEFEED;
                        stext += tr ( "Coded: " );
                        stext += tr ( "yes" );
                        prog_data.description += stext;
                    }
                    if ( prog_data.age_marker )
                    {
                        stext = LINEFEED;
                        stext += tr ( "Age protected: " );
                        stext += tr ( "yes" );
                        prog_data.description += stext;
                    }
                    if ( prog_data.dolby )
                    {
                        stext = LINEFEED;
                        stext += tr ( "Dolby surround: " );
                        stext += tr ( "yes" );
                        prog_data.description += stext;
                    }
                    if ( prog_data.stereo )
                    {
                        stext = LINEFEED;
                        stext += tr ( "Stereo: " );
                        stext += tr ( "yes" );
                        prog_data.description += stext;
                    }
                    if ( prog_data.sequence )
                    {
                        stext = LINEFEED;
                        stext += tr ( "Sequence: " );
                        stext += itoa ( prog_data.sequence );
                        prog_data.description += stext;
                    }
                    if ( prog_data.wide )
                    {
                        stext = LINEFEED;
                        stext += tr ( "Wide: " );
                        stext += tr ( "yes" );
                        prog_data.description += stext;
                    }
                    if ( prog_data.primetime )
                    {
                        stext = LINEFEED;
                        stext += tr ( "Prime time: " );
                        stext += tr ( "yes" );
                        prog_data.description += stext;
                    }
                    prog_data.description += LINEFEED;
                    prog_data.description +=  tr ( "Genre: " );
                    prog_data.description += get_genre ( prog_data.genre );
                    prog_data.description += LINEFEED;

                    //FIXME image processing causes memleak!
                    if ( prog_data.jpeg.size() )
                    {
                        // Construct path for EpgDataFileName
                        string jpath = EpgDir;
                        jpath += "/";
                        jpath += prog_data.jpeg.c_str();

                        // convert to png -> adjust filename for output
                        string sout;
                        string sout_raw;

                        sout = jpath.c_str();
                        sout.erase ( jpath.size() - 4 );
                        sout_raw = sout.c_str();

                        sout += ".png";
                        sout_raw += ".raw";

                        if ( !file_exists ( ( char * ) ( sout+".converted" ).c_str() ) && file_exists ( ( char * ) jpath.c_str() ) )
                        {
//                          cPngUtils *img = new cPngUtils ( 32000, 32000 );  // Images max. size 32000 x 32000 pixel

                            int rc = img.ReadJpeg ( ( char * ) jpath.c_str() );
                            if ( rc == 0 )
                            {
                                char buf[512];
                                sprintf ( buf, "%s: %s => %s", tr ( "image converted" ), jpath.c_str(), sout.c_str() );
                                do_log ( EpgDir.c_str(), LOGFILE_IMPORT, buf );

                                if ( iepgSetup.convert_to_png )
                                {
                                    if ( !iepgSetup.pics_only_primetime || ( iepgSetup.pics_only_primetime && prog_data.primetime ) )
                                    {
                                        rc = img.WriteFile ( 250, 0 /* xpos */, 0 /* ypos */, ( char * ) sout.c_str(), false /* bIsPNG */ );
                                        if ( rc != 0 )
                                            esyslog ( "iepg: write %s ended, Status=%d", ( char * ) sout.c_str(), rc );
                                    }
                                }
                                if ( iepgSetup.convert_to_raw )
                                    img.DumpRaw ( ( char * ) sout_raw.c_str(), 2 ); // Dump Content of image (scaled by 50%) in raw format for performance reasons

                                // write file image path + .converted to indicate that this image is already converted and there is no need to do it again
                                FILE * marker;
                                marker = fopen ( ( char * ) ( sout + ".converted" ).c_str(), "w" );
                                fclose ( marker );

//                            delete img;
                            }
                            else
                                esyslog ( "iepg: jpeg read %s ended, Status=%d", ( char * ) jpath.c_str(), rc );
                        }
//                        else
//                            dsyslog ( "iepg: file %s already converted", ( char * ) sout.c_str() );

                        // save filename of png image
                        prog_data.png = "";
                        prog_data.png += sout.c_str();

                    }

                    /* ch_data-> */
                    ProgrammeList.push_back ( prog_data );
//                    dsyslog ( "iepg: epg data for channelid %d: %s", prog_data.ChannelID.Source(), prog_data.title.c_str() );


                    prog_data.png = "";  // reset png path for next entry
                    prog_data.eventid = 0L;
                }

            }
        }

    }

    delete line; // free memory

    fclose ( xmlFile );
    return 1;
}


