/*
*  C Implementation: xmlepg2vdrthread
*
* Description:  A thread is created that runs in the background collecting
*               xml-EPG data from websites and loading it into vdr schedules
*
* Author: RealQuick <RealQuick@doubleQuick.de>, (C) 2009
*
* Copyright: See COPYING file that comes with this distribution
*
*/
#include <vdr/thread.h>
#include <curl/curl.h>
#include <math.h>
#include <stdio.h>
#include <sys/types.h>
#include <dirent.h>
#include "setup.h"
#include "channelData.h"
#include <vdr/channels.h>       // for Channels
#include "common.h"
#include "svdrpc.h"

#include "xmlepg2vdrthread.h"
#include "misc.h"

extern std::vector < cProviderStatus > Providers;
extern int *Reverse_ChannelList;

cThreadXmlEpg2vdr::cThreadXmlEpg2vdr ( char *xConfigDir, bool xforceEPG, bool xforcePIC, bool xCleanEpgOnly, bool xQuiet) :cThread ( "XmlEpg2vdr" )
{
    isAlreadyRunning = false;
    WorkingDay = 0;   // initialize to 0 (means: download and load epg for the actual day)

    sConfigDir = xConfigDir;

    forceEPG = xforceEPG;
    forcePIC = xforcePIC;

    bQuiet = xQuiet;

    CleanEpgOnly = xCleanEpgOnly;
}

cThreadXmlEpg2vdr::~cThreadXmlEpg2vdr ()
{
    if ( isAlreadyRunning )
    {
        dsyslog ( "iepg: destructor of cThreadXmlEpg2vdr -> already running -> Cancel(10)" );
        isAlreadyRunning = false;
        do_log ( iepgSetup.sDataDirectory.c_str(), LOGFILE_THREAD, ( char * ) tr ( "iepg is already running"), false);
        do_log ( iepgSetup.sDataDirectory.c_str(), LOGFILE_STATUS, ( char * ) tr ( "errors" ), true);

        Cancel ( 10 );
    }
}

void cThreadXmlEpg2vdr::killThread ( int sec )
{
    dsyslog ( "iepg: KillThread" );

    Cancel ( sec );
}

/********** download URL from network *********/
int cThreadXmlEpg2vdr::downloadURL ( char *URL, char *filename )
{
    CURL *curl;
    CURLcode res;
    int returnFlag = 1;

    curl = curl_easy_init ();
    if ( curl )
    {
        FILE *fp = fopen ( filename, "w" ); // should overwrite file
        curl_easy_setopt ( curl, CURLOPT_URL, URL );
        curl_easy_setopt ( curl, CURLOPT_USERAGENT, USERAGENTSTRING );
        curl_easy_setopt ( curl, CURLOPT_WRITEDATA, fp );
        res = curl_easy_perform ( curl );
        if ( res )
        {
            returnFlag = 0;
        }

        fclose ( fp );
    }
    /* always cleanup */
    curl_easy_cleanup ( curl );

    if (get_file_size(filename) < 200)
    {
        esyslog("iepg: Error downloading epg data. Please check your PIN and download URL");
        do_log ( iepgSetup.sDataDirectory.c_str(), LOGFILE_THREAD, ( char * ) tr ( "Error downloading epg data. Please check your PIN and download URL"), false);

        do_log ( iepgSetup.sDataDirectory.c_str(), LOGFILE_STATUS, ( char * ) tr ( "errors" ), true);
        return false;
    }

    return returnFlag;
}

void cThreadXmlEpg2vdr::SetQuiet(bool xQuiet)
{
    bQuiet = xQuiet;
}

void cThreadXmlEpg2vdr::Action ()
{
    if ( isAlreadyRunning )
        return;
    isAlreadyRunning = true;

    switch (iepgSetup.Priority)
    {
    case 0:
        SetPriority(1);
        break;
    case 1:
        SetPriority(10);
        break;
    case 2:
        SetPriority(15);
        break;
    case 3:
        SetPriority(19);
        break;
    }

    // Just Clean the schedule from duplicate events and exit
    // Usually called from ChanneSwitch event (status monitor)
    if (CleanEpgOnly)
    {
        Lock();
        sleep(2);

        cCleanChannelData Metzger(NULL, tChannelID() /* chn->GetChannelID() */, Reverse_ChannelList);	// TODO Reverse_ChannelList is global :)
        Metzger.CleanNow(false);

        Unlock();
        isAlreadyRunning = false;

        // reset to default values
        forceEPG = false;
        forcePIC = false;
        CleanEpgOnly = false;

        return;
    }

#if 0
    // Wait until scheduled time is reached
    while ( Running() && time ( NULL ) < t_Start )
        sleep ( 10 );

    // if startup method = at specific time set next import time immediately
    if (iepgSetup.hour_to_start > IMPORT_AFTER_STARTUP)
    {
        t_Start + ( time_t ) ( 24 * 60 * 60);

        struct tm *ts;
        ts = localtime(&t_Start);
        dsyslog("iepg: set next startup to %s", asctime(ts));
    }
#endif

    iepgStartMessage ( MAIN_PROG, tr ( "started" ) );
//     iepgStartMessage(PIC_UPDATE, tr("started"));
//     iepgStartMessage(EPG_UPDATE, tr("started"));

    Lock ();

    bWarnings = false;  // Reset warning flag

//            Skins.Message ( mtInfo, tr ( "Cleaning up" ), 1 );

    int nCount;
    char buf[512];

    do_log ( iepgSetup.sDataDirectory.c_str(), LOGFILE_THREAD, ( char * ) tr ( "import of epg data from internet is starting" ), true );
    do_log ( iepgSetup.sDataDirectory.c_str(), LOGFILE_STATUS, ( char * ) tr ( "pending" ), true );


    // Cleanup raw Pictures
    string sImageDir = iepgSetup.sImageDirectory;
    nCount = Cleanup ( ( char * ) sImageDir.c_str(), ".raw" );
    sprintf ( buf, "%s => %d %s", tr ( "remove outdated raw pictures" ), nCount,  tr ( "removed" ) );
    do_log ( iepgSetup.sDataDirectory.c_str(), LOGFILE_THREAD, buf );

    // Cleanup PNG Pictures
    nCount = Cleanup ( ( char * ) sImageDir.c_str(), ".png" );
    sprintf ( buf, "%s => %d %s", tr ( "remove outdated png pictures" ), nCount,  tr ( "removed" ) );
    do_log ( iepgSetup.sDataDirectory.c_str(), LOGFILE_THREAD, buf );

    // Cleanup Directories for all providers above 0
    nCount = 0;
    for(unsigned int i=1;i<=iepgSetup.SetupProviders.size();i++)
    {
       sprintf ( buf, "%s/%d", iepgSetup.sDataDirectory.c_str(), i); 
       nCount += Cleanup ( buf, "." );
    }

    sprintf ( buf, "%s => %d %s", tr ( "remove outdated folders" ), nCount,  tr ( "removed" ) );
    do_log ( iepgSetup.sDataDirectory.c_str(), LOGFILE_THREAD, buf );


    do_log ( iepgSetup.sDataDirectory.c_str(), LOGFILE_THREAD, ( char * ) tr ( "epg download is starting" ) );

    if(!bQuiet)
        Skins.QueueMessage ( mtInfo, tr ( "EPG data import is starting" ), 4 );

    // this function runs in the background as a thread
    // cThread::Start() call this function

    /*
     * at times Action() was called twice, cause unknow.
     *
     * So, checks isAlreadyRunning and Lock() in place
     * */

    cChannelData *channelData = NULL;

    std::vector < cURLFile > Urls;
    cURLFile tmp_obj;

    DIR *tmpdir;
    string LastEpgFileName = "";
    int prgCount = 0;

    // Create base directory for download
    if ( ( tmpdir=opendir ( iepgSetup.sDataDirectory.c_str() ) ) == NULL )
    {
        isyslog ( "Creating %s\n", iepgSetup.sDataDirectory.c_str() );
        if ( mkdir ( iepgSetup.sDataDirectory.c_str(), ACCESSPERMS ) )
        {
            esyslog ( "Creation of %s failed.\n", iepgSetup.sDataDirectory.c_str() );
            isAlreadyRunning = false;
            do_log ( iepgSetup.sDataDirectory.c_str(), LOGFILE_THREAD, ( char * ) tr ( "Creation of folder failed"), false);
            do_log ( iepgSetup.sDataDirectory.c_str(), LOGFILE_STATUS, ( char * ) tr ( "errors" ), true );
            return;
        }
    }
    else
    {
        closedir ( tmpdir );
    }

    dsyslog ( "iepg: %d providers defined", iepgSetup.SetupProviders.size() );

    ChannelList = new cProviderChannelList[MAX_PROVIDER_CHANNEL + PROVIDER_HASHSIZE + COLLISION_TABSIZE];
    for (int p=0;p<(MAX_PROVIDER_CHANNEL + PROVIDER_HASHSIZE + COLLISION_TABSIZE);p++) ChannelList[p].ChannelId = tChannelID();
//    ProviderChannelNames = new string[MAX_PROVIDER_CHANNEL + PROVIDER_HASHSIZE + COLLISION_TABSIZE];

    char *file;
    char sDir[256];
    char *d;
    bool bUpdate = false;

    std::vector < cURLFile >::iterator iter = Urls.begin ();	// iterator for looping through url list


    for ( unsigned int nProvider=0;Running() && nProvider<iepgSetup.SetupProviders.size();nProvider++ )
    {
        prgCount = 0;

        WorkingDay = 0;   // initialize to 0 (means: download and load epg for the actual day)

//         SetupStore ("LastUpdate", time(NULL));

        Urls.clear();

        // Loop for 0 .. days in advanced (if possible)
        bool bLoop = true; // if days in advance for provider is not possible -> bLoop = false
        while ( Running() && bLoop && WorkingDay <= iepgSetup.SetupProviders[nProvider].days )
        {

            iepgMessage ( MAIN_PROG, ( 100 * 2 * WorkingDay  + EPG_LOAD ) / ( 3 /* Steps per day*/ * iepgSetup.SetupProviders[nProvider].days ), tr ( "downloading" ) );


            dsyslog ( "iepg: provider %d autoupdate=%d", nProvider, iepgSetup.SetupProviders[nProvider].autoupdate );

            if ( iepgSetup.SetupProviders[nProvider].autoupdate )
            {
// Dont clear Urls
//                Urls.clear();

		bDownloadError = false;
                if ( !ReadUrlList ( this, &Urls, sConfigDir, nProvider, ChannelList, /* Download=*/true, WorkingDay, &bLoop ) )
                {
//                    Unlock ();
//                    isAlreadyRunning = false;
                    do_log ( iepgSetup.sDataDirectory.c_str(), LOGFILE_THREAD, ( char * ) tr ( "Download of epgdata failed"), false);
                    do_log ( iepgSetup.sDataDirectory.c_str(), LOGFILE_STATUS, ( char * ) tr ( "warning" ), true);

		    bDownloadError = true; // Do not cancel thread. just set error flag
		    bWarnings = true;
//                    return;
                } else
		{
                    isyslog ( "iepg: EPG download from internet complete" );
		}
            }
            else
            {
                bLoop = false; // No autoupdate activated -> leave loop
            }
            dsyslog("iepg: Provider %d, Providertype=%d",nProvider, iepgSetup.SetupProviders[nProvider].ProviderType);

            if ( bLoop )
            {
                switch ( bDownloadError ? -1 : iepgSetup.SetupProviders[nProvider].ProviderType )  // DFo not import data in case of an download error
                {
                case 1:	// Do EPG update for provider type 1 (epgdata.com)
                    sprintf ( sDir, "%s/%d", iepgSetup.sDataDirectory.c_str(), nProvider );
                    d = get_dir_per_day ( sDir, WorkingDay );
                    strcpy ( sDir, d );
                    free ( d );

                    do_log ( sDir, LOGFILE_STATUS, ( char * ) tr ( "pending" ), true );


                    file = get_xml_from_zip ( sDir );
                    dsyslog("iepg: File from %s to load %s", sDir, file);
                    bUpdate = false;

                    // forceEPG = false -> check for xml_processed
                    // xml file must exists!
                    if ( file )
                        if ( !forceEPG )
                        {
                            // if xml already processed -> skipping it
                            string chkfile = sDir;
                            chkfile += "/";
                            chkfile += file;
                            chkfile += "_processed";

                            if ( file_exists ( ( char * ) chkfile.c_str() ) )
                            {
                                dsyslog ( "iepg: file %s was already processed, Skipping it", file );
                                do_log ( sDir, LOGFILE_IMPORT, ( char * ) tr ( "file was already processed, Skipping it" ) );

                                do_log ( iepgSetup.sDataDirectory.c_str(), LOGFILE_THREAD, ( char * ) tr ( "XML file was already processed, Skipping it" ) );
                                do_log ( sDir, LOGFILE_STATUS, ( char * ) tr ( "skipped" ), true );

                            }
                            else
                                bUpdate = true;
                        }
                        else
                            bUpdate = true;
                    if ( bUpdate )
                    {

                        iepgMessage ( MAIN_PROG, 100 * ( 2 * WorkingDay  + EPG_UPDATE ) / ( 3 /* Steps per day*/ * iepgSetup.SetupProviders[nProvider].days + EPG_LOAD ), tr ( "loading" ) );

                        /* load EPG data into schedule */
                        channelData = new cChannelData ( sDir, file, 1 ); // TODO Check for errors

                        // Reads EPGDATAFILE, Parses and adds EPG data to channel "channelName_str"

                        iepgMessage ( MAIN_PROG, 100 * ( 2 * WorkingDay + EPG_UPDATE ) / ( 3 /* Steps per day*/ * iepgSetup.SetupProviders[nProvider].days ), tr ( "updating" ) );

                        prgCount = channelData->loadEpgData ( ChannelList );
                        if ( prgCount<0 )
                        {
                            esyslog ( "iepg: Error loading/parsing EPG data for xml file %s/%s",  sDir, file );
			    bWarnings = true;

                            do_log ( sDir, LOGFILE_STATUS, ( char * ) tr ( "errors" ), true );
			    char buff[512];
			    sprintf(buff, tr ( "Error loading/parsing EPG data for xml file %s/%s" ),  sDir, file);
                            do_log ( iepgSetup.sDataDirectory.c_str(), LOGFILE_THREAD, buff );
                        }
                        else
                        {
                            char sMeld[256];
                            sprintf ( sMeld, tr("iepg: Sucessfully loaded EPG for channel xml file %s/%s. Number of programs = %d"), sDir, file , prgCount );
                            dsyslog ( sMeld );
                            do_log ( iepgSetup.sDataDirectory.c_str(), LOGFILE_THREAD, sMeld );


                            // rename xml file to mark as processed
                            char sMoveFile[512];
                            sprintf ( sMoveFile, "mv %s/%s %s/%s_processed", sDir, file, sDir, file );
                            dsyslog ( "iepg: rename xml file to mark as processed: %s", sMoveFile );
                            SystemExec ( sMoveFile );

                            char buf[512];
                            sprintf ( buf, "%s %s %s = %d", tr ( "Successfully loaded EPG from xml" ), file, tr ( "Number of programs" ), prgCount );
                            do_log ( sDir, LOGFILE_IMPORT, buf );

                            do_log ( sDir, LOGFILE_STATUS, ( char * ) tr ( "successful" ), true );
                        }
                    }

                    if ( file )
                        free ( file );

                    if ( channelData != NULL )
                    {
                        delete channelData;
                        channelData = NULL;
                    }
                    break;

                case 0:   // Do update of epg data (provider type 0 bleb.org)
                    isyslog ( "iepg: Begin loading EPG data for provider type 0" );

                    LastEpgFileName = "";

//                        char sDir[256];

                    dsyslog("iepg: about to loop through URLs %d times", Urls.size());

                    for (iter = Urls.begin() ; Running() && iter != Urls.end (); iter++ )
                    {

                        if ( iter->isError )
                        {
                            esyslog("iepg: do not load epg data. cause: errors");
                            continue;               // donot load channel data if error in downloading URL
                        }
                        sprintf ( sDir, "%s/%d", iepgSetup.sDataDirectory.c_str(), iter->ProviderType );
                        dsyslog("iepg: Loadepg with file %s", iter->EpgFileName.c_str ());
                        /* load EPG data into schedule */
                        {
                            channelData = new cChannelData ( sDir, iter->EpgFileName.c_str (), iter->ProviderType ); // TODO Check for errors
                        }
                        // Reads EPGDATAFILE, Parses and adds EPG data to channel "channelName_str"
                        if ( strlen ( iter->ChannelName.c_str () ) > 0 )
                        {
//                                prgCount = channelData->loadEpgData ( iter->ChannelName.c_str (), ( char * ) iter->ChannelId.c_str () );
                            dsyslog("iepg: Loadepg for channel %s, ChannelId=%s", iter->ChannelName.c_str (), iter->ChannelId.c_str());
                            prgCount = channelData->loadEpgData ( ChannelList, atoi(iter->ChannelId.c_str()));
                            if ( prgCount<0 )
                            {
                                esyslog ( "iepg: Error loading/parsing EPG data for channel %s",iter->ChannelName.c_str() );
			        bWarnings = true;
                                char buf[512];
                                sprintf ( buf, tr("Error loading/parsing EPG data for channel %s"), iter->ChannelName.c_str());
                                 do_log ( iepgSetup.sDataDirectory.c_str(), LOGFILE_THREAD, buf );
                            }
                            else
                            {
                                dsyslog ( "iepg: Sucessfully loaded EPG for channel %s. Number of programmes = %d",iter->ChannelName.c_str() , prgCount );

                                char buf[512];
                                sprintf ( buf, "%s %s %s = %d", tr ( "Sucessfully loaded EPG for channel" ), iter->ChannelName.c_str(), tr ( "Number of programmes" ), prgCount );
                                do_log ( iepgSetup.sDataDirectory.c_str(), LOGFILE_THREAD, buf );

                            }
                        }

                    } // end loop Urls

//            }
                    break;
                default:
                    break;
                }
            }
	    // Import was processed for the current provider
            if (iepgSetup.SetupProviders[nProvider].autoupdate)
            {
                dsyslog ( "iepg: finished loading EPG data" );
            
                if ( prgCount && !bDownloadError)
                {
                    dsyslog ( "iepg: data import successful" );
                }
                else
                {
		    // Data should been updated
		    if(bUpdate)
		    {
                        esyslog ( "iepg: data import failed (download error)" );
                        bWarnings = true;
		    }
                }
            }
            if ( channelData != NULL )
            {
                delete channelData;
                channelData = NULL;
            }

            WorkingDay++; // start cycle for next day
        }  // loop for days in advance (not supported by provider type 0)
    }   // main loop for provider

    Urls.clear();

    if ( channelData != NULL )
    {
        delete channelData;
        channelData = NULL;
    }
    delete [] ChannelList;
//
    cSchedules::Cleanup ( true );
    dsyslog ( "iepg: Cleanup(true) forced" );
    do_log ( iepgSetup.sDataDirectory.c_str(), LOGFILE_THREAD, ( char * ) tr ( "EPG data saved to disk" ) );


    iepgEndMessage ( MAIN_PROG, tr ( "Finished" ) );
//     iepgEndMessage(PIC_UPDATE, tr("Finished"));
//     iepgEndMessage(EPG_UPDATE, tr("Finished"));

    // save epg to disk immediately to avoid loss of data after reboot/standby
#ifdef RBLITE	// no need for the AVG
    SystemExec ( "cp /tmp/epg.data /media/hd/epg.data" );
    dsyslog ( "iepg: saved epg to /media/hd/epg.data" );

    do_log ( iepgSetup.sDataDirectory.c_str(), LOGFILE_THREAD, ( char * ) tr ( "saved epg to /media/hd/epg.data" ) );
#endif
    do_log ( iepgSetup.sDataDirectory.c_str(), LOGFILE_THREAD, ( char * ) tr ( "epg import from internet finished" ) );

    if(bWarnings)
    	do_log ( iepgSetup.sDataDirectory.c_str(), LOGFILE_STATUS, ( char * ) tr ( "warning" ), true );
    else
    	do_log ( iepgSetup.sDataDirectory.c_str(), LOGFILE_STATUS, ( char * ) tr ( "successful" ), true );


    Unlock ();
    isAlreadyRunning = false;

    if(!bQuiet)
    {
        Skins.QueueMessage ( mtInfo, tr ( "EPG data successfully imported from Internet" ), 4 );
	bQuiet = true;  // Reset to default
    }

    // Reset flags for next call
    forceEPG = false;
    forcePIC = false;

    return;
}                             // end Action()

